from __future__ import annotations
from pydantic import BaseModel
from functools import partial
from types import MappingProxyType
from typing import Optional, Tuple
import pandas as pd

# local imports
from . import utilities
from .type_definitions import GeographicScale, Year


class FieldNameMap(BaseModel):
    """Map from canonical hydrotools SVI field names to SVI field names from another provenance."""

    state_name: str  # State name
    state_abbreviation: str  # State abbreviation
    county_name: str  # County name

    state_fips: str  # State FIPS code
    county_fips: str  # County name
    fips: str  # Census tract or county fips code

    svi_edition: str  # year corresponding to svi release (this assumes 2 SVI's will not be release in a given year in the future)

    socioeconomic_rank: str  # theme_1_rank : Socioeconomic
    household_comp_and_disability_rank: str  # theme_2_rank : Household Composition / Disability
    minority_status_and_lang_rank: str  # theme_3_rank : Minority Status / Language
    housing_type_and_trans_rank: str  # theme_4_rank : Housing Type / Transportation
    svi_rank: str  # aggregated overall percentile ranking

    socioeconomic_value: Optional[str]  # theme_1_value : Socioeconomic
    household_comp_and_disability_value: Optional[
        str
    ]  # theme_2_value : Household Composition / Disability
    minority_status_and_lang_value: Optional[
        str
    ]  # theme_3_value : Minority Status / Language
    housing_type_and_trans_value: Optional[
        str
    ]  # theme_4_value : Housing Type / Transportation

    # aggregated overall value; sum of values from themes 1, 2, 3, 4.
    svi_value: Optional[str]

    @staticmethod
    def create_missing_fields(df: pd.DataFrame) -> pd.DataFrame:
        """subclasses should override this to create missing dataframe fields from existing fields.
        default behavior is to return input df.
        """
        return df


### Mapping types for data sourced from: https://services3.arcgis.com/ZvidGQkLaDJxRSJ2/ArcGIS/rest/services/. ###

## SVI Counties Types ##

# 2000s data does not include theme values
CdcEsri2000CountiesFieldNameMap = FieldNameMap(
    state_name="STATE_NAME",
    state_abbreviation="STATE_ABBR",
    county_name="COUNTY",
    state_fips="STATE_FIPS",
    county_fips="CNTY_FIPS",
    fips="STCOFIPS",
    svi_edition="2000",
    socioeconomic_rank="USG1TP",
    household_comp_and_disability_rank="USG2TP",
    minority_status_and_lang_rank="USG3TP",
    housing_type_and_trans_rank="USG4TP",
    svi_rank="USTP",
    # below fields do not exits in 2000s data product
    socioeconomic_value=None,
    household_comp_and_disability_value=None,
    minority_status_and_lang_value=None,
    housing_type_and_trans_value=None,
    svi_value=None,
)

CdcEsri2010CountiesFieldNameMap = FieldNameMap(
    state_name="FIRST_STATE_NAME",
    state_abbreviation="FIRST_STATE_ABBR",
    county_name="FIRST_COUNTY",
    state_fips="FIRST_STATE_FIPS",
    county_fips="FIRST_CNTY_FIPS",
    fips="STCOFIPS",
    svi_edition="2010",
    socioeconomic_rank="R_PL_THEME1",
    household_comp_and_disability_rank="R_PL_THEME2",
    minority_status_and_lang_rank="R_PL_THEME3",
    housing_type_and_trans_rank="R_PL_THEME4",
    svi_rank="R_PL_THEMES",
    socioeconomic_value="S_PL_THEME1",
    household_comp_and_disability_value="S_PL_THEME2",
    minority_status_and_lang_value="S_PL_THEME3",
    housing_type_and_trans_value="S_PL_THEME4",
    svi_value="S_PL_THEMES",
)


class _CdcEsriMissingCountyFipsFieldNameMap(FieldNameMap):
    @staticmethod
    def create_missing_fields(df: pd.DataFrame) -> pd.DataFrame:
        """
        derive `county_fips` from `fips`. county fips codes are five digits long, where the first
        two digits are the state fips.
        source: https://transition.fcc.gov/oet/info/maps/census/fips/fips.txt
        """
        return df.assign(county_fips=lambda d: d["fips"].str.slice(2, 5))


# svi_edition is excluded, so it can be parametrized
_CdcEsriCountiesFieldNameMap = partial(
    _CdcEsriMissingCountyFipsFieldNameMap,
    state_name="STATE",
    state_abbreviation="ST_ABBR",
    county_name="COUNTY",
    state_fips="ST",
    county_fips="FIPS",  # calculated FIPS[2:5]
    fips="FIPS",
    socioeconomic_rank="RPL_THEME1",
    household_comp_and_disability_rank="RPL_THEME2",
    minority_status_and_lang_rank="RPL_THEME3",
    housing_type_and_trans_rank="RPL_THEME4",
    svi_rank="RPL_THEMES",
    socioeconomic_value="SPL_THEME1",
    household_comp_and_disability_value="SPL_THEME2",
    minority_status_and_lang_value="SPL_THEME3",
    housing_type_and_trans_value="SPL_THEME4",
    svi_value="SPL_THEMES",
)

CdcEsri2014CountiesFieldNameMap = _CdcEsriCountiesFieldNameMap(svi_edition="2014")
CdcEsri2016CountiesFieldNameMap = _CdcEsriCountiesFieldNameMap(svi_edition="2016")
CdcEsri2018CountiesFieldNameMap = _CdcEsriCountiesFieldNameMap(svi_edition="2018")


##  SVI Tract Types  ##

# 2000s data does not include theme values
CdcEsri2000TractsFieldNameMap = FieldNameMap(
    state_name="STATE_NAME",
    state_abbreviation="STATE_ABBR",
    county_name="COUNTY",
    state_fips="STATE_FIPS",
    county_fips="CNTY_FIPS",
    fips="FIPS",
    svi_edition="2000",
    socioeconomic_rank="USG1TP",
    household_comp_and_disability_rank="USG2TP",
    minority_status_and_lang_rank="USG3TP",
    housing_type_and_trans_rank="USG4TP",
    svi_rank="USTP",
    # below fields do not exits in 2000s data product
    socioeconomic_value=None,
    household_comp_and_disability_value=None,
    minority_status_and_lang_value=None,
    housing_type_and_trans_value=None,
    svi_value=None,
)

CdcEsri2010TractsFieldNameMap = FieldNameMap(
    state_name="STATE_NAME",
    state_abbreviation="STATE_ABBR",
    county_name="COUNTY",
    state_fips="STATE_FIPS",
    county_fips="CNTY_FIPS",
    fips="FIPS",
    svi_edition="2010",
    socioeconomic_rank="R_PL_THEME1",
    household_comp_and_disability_rank="R_PL_THEME2",
    minority_status_and_lang_rank="R_PL_THEME3",
    housing_type_and_trans_rank="R_PL_THEME4",
    svi_rank="R_PL_THEMES",
    socioeconomic_value="S_PL_THEME1",
    household_comp_and_disability_value="S_PL_THEME2",
    minority_status_and_lang_value="S_PL_THEME3",
    housing_type_and_trans_value="S_PL_THEME4",
    svi_value="S_PL_THEMES",
)

# svi_edition is excluded, so it can be parametrized
_CdcEsriTractFieldNameMap = partial(
    _CdcEsriMissingCountyFipsFieldNameMap,
    state_name="STATE",
    state_abbreviation="ST_ABBR",
    county_name="COUNTY",
    state_fips="ST",
    county_fips="STCNTY",  # calculated FIPS[2:5]
    fips="FIPS",
    socioeconomic_rank="RPL_THEME1",
    household_comp_and_disability_rank="RPL_THEME2",
    minority_status_and_lang_rank="RPL_THEME3",
    housing_type_and_trans_rank="RPL_THEME4",
    svi_rank="RPL_THEMES",
    socioeconomic_value="SPL_THEME1",
    household_comp_and_disability_value="SPL_THEME2",
    minority_status_and_lang_value="SPL_THEME3",
    housing_type_and_trans_value="SPL_THEME4",
    svi_value="SPL_THEMES",
)

CdcEsri2014TractsFieldNameMap = _CdcEsriTractFieldNameMap(svi_edition="2014")
CdcEsri2016TractsFieldNameMap = _CdcEsriTractFieldNameMap(svi_edition="2016")
CdcEsri2018TractsFieldNameMap = _CdcEsriTractFieldNameMap(svi_edition="2018")

_CdcEsriFieldNameMapFactory: MappingProxyType[
    Tuple[str, str], FieldNameMap
] = MappingProxyType(
    {
        # year, geographic scale: FieldNameMap
        # counties
        ("2000", "county"): CdcEsri2000CountiesFieldNameMap,
        ("2010", "county"): CdcEsri2010CountiesFieldNameMap,
        ("2014", "county"): CdcEsri2014CountiesFieldNameMap,
        ("2016", "county"): CdcEsri2016CountiesFieldNameMap,
        ("2018", "county"): CdcEsri2018CountiesFieldNameMap,
        # tracts
        ("2000", "census_tract"): CdcEsri2000TractsFieldNameMap,
        ("2010", "census_tract"): CdcEsri2010TractsFieldNameMap,
        ("2014", "census_tract"): CdcEsri2014TractsFieldNameMap,
        ("2016", "census_tract"): CdcEsri2016TractsFieldNameMap,
        ("2018", "census_tract"): CdcEsri2018TractsFieldNameMap,
    }
)


def CdcEsriFieldNameMapFactory(
    geographic_scale: GeographicScale, year: Year
) -> FieldNameMap:
    geographic_scale = utilities.validate_geographic_scale(geographic_scale)  # type: ignore
    year = utilities.validate_year(year)  # type: ignore

    search_tuple = (year, geographic_scale)
    return _CdcEsriFieldNameMapFactory[search_tuple]  # type: ignore


# Developer notes

# state_fips:
#     counties: STATE_FIPS, FIRST_STATE_FIPS, ST, ST, ST
#     tracts: STATE_FIPS, STATE_FIPS, ST, ST, ST
# county_fips:
#     counties: CNTY_FIPS, FIRST_CNTY_FIPS, FIPS[len(ST):], FIPS[len(ST):], FIPS[len(ST):]
#     tracts: CNTY_FIPS, CNTY_FIPS, STCNTY[len(ST):], STCNTY[len(ST):], STCNTY[len(ST):]
# fips:
#     counties: STCOFIPS, STCOFIPS, FIPS, FIPS, FIPS
#     tracts: FIPS, FIPS, FIPS, FIPS, FIPS
# state_name:
#     counties: STATE_NAME, FIRST_STATE_NAME, STATE, STATE, STATE
#     tracts: STATE_NAME, STATE_NAME, STATE.strip(), STATE, STATE
# state_abbreviation:
#     counties: STATE_ABBR, FIRST_STATE_ABBR, ST_ABBR, ST_ABBR, ST_ABBR
#     tracts: STATE_ABBR, STATE_ABBR, ST_ABBR, ST_ABBR, ST_ABBR
# county_name:
#     counties: COUNTY, FIRST_COUNTY, COUNTY, COUNTY, COUNTY
#     tracts: COUNTY, COUNTY, COUNTY.strip(), COUNTY, COUNTY
# svi_edition:
#     counties: 2000, 2010, 2014, 2016, 2018
#     tracts: 2000, 2010, 2014, 2016, 2018
# rank_theme_1
#     counties: USG1TP, R_PL_THEME1, RPL_THEME1, RPL_THEME1, RPL_THEME1
#     tracts: USG1TP, R_PL_THEME1, RPL_THEME1, RPL_THEME1, RPL_THEME1
# rank_theme_2
#     counties: USG2TP, R_PL_THEME2, RPL_THEME2, RPL_THEME2, RPL_THEME2
#     tracts: USG2TP, R_PL_THEME2, RPL_THEME2, RPL_THEME2, RPL_THEME2
# rank_theme_3
#     counties: USG3TP, R_PL_THEME3, RPL_THEME3, RPL_THEME3, RPL_THEME3
#     tracts: USG3TP, R_PL_THEME3, RPL_THEME3, RPL_THEME3, RPL_THEMES3
# rank_theme_4
#     counties: USG3TP, R_PL_THEME4, RPL_THEME4, RPL_THEME4, RPL_THEME4
#     tracts: USG3TP, R_PL_THEME4, RPL_THEME4, RPL_THEME4, RPL_THEME4
# rank_svi
#     counties: USTP, R_PL_THEMES, RPL_THEMES, RPL_THEMES, RPL_THEMES
#     tracts: USTP, R_PL_THEMES, RPL_THEMES, RPL_THEMES, RPL_THEMES
# value_theme_1
#     counties: NA, S_PL_THEME1, SPL_THEME1, SPL_THEME1, SPL_THEME1
#     tracts: NA, S_PL_THEME1, SPL_THEME1, SPL_THEME1, SPL_THEME1
# value_theme_2
#     counties: NA, S_PL_THEME2, SPL_THEME2, SPL_THEME2, SPL_THEME2
#     tracts: NA, S_PL_THEME2, SPL_THEME2, SPL_THEME2, SPL_THEME2
# value_theme_3
#     counties: NA, S_PL_THEME3, SPL_THEME3, SPL_THEME3, SPL_THEME3
#     tracts: NA, S_PL_THEME3, SPL_THEME3, SPL_THEMES3, SPL_THEME3
# value_theme_4
#     counties: NA, S_PL_THEME4, SPL_THEME4, SPL_THEME4, SPL_THEME4
#     tracts: NA, S_PL_THEME4, SPL_THEME4, SPL_THEME4, SPL_THEME4
# value_svi
#     counties: NA, S_PL_THEMES, SPL_THEMES, SPL_THEMES, SPL_THEMES
#     tracts: NA, S_PL_THEMES, SPL_THEMES, SPL_THEMES, SPL_THEMES



import pytest
from hydrotools.svi_client import SVIClient

LOCATIONS = (
    "al",
    "ak",
    "az",
    "ar",
    "ca",
    "co",
    "ct",
    "de",
    "dc",
    "fl",
    "ga",
    "hi",
    "id",
    "il",
    "in",
    "ia",
    "ks",
    "ky",
    "la",
    "me",
    "md",
    "ma",
    "mi",
    "mn",
    "ms",
    "mo",
    "mt",
    "ne",
    "nv",
    "nh",
    "nj",
    "nm",
    "ny",
    "nc",
    "nd",
    "oh",
    "ok",
    "or",
    "pa",
    "ri",
    "sc",
    "sd",
    "tn",
    "tx",
    "ut",
    "vt",
    "va",
    "wa",
    "wv",
    "wi",
    "wy",
)

YEARS = ("2000", "2010", "2014", "2016", "2018")
GEOGRAPHIC_SCALES = ("county", "census_tract")


@pytest.mark.slow
@pytest.mark.parametrize("location", LOCATIONS)
@pytest.mark.parametrize("year", YEARS)
@pytest.mark.parametrize("scale", GEOGRAPHIC_SCALES)
def test_svi_client_get_integration(location, year, scale, loop):
    client = SVIClient(enable_cache=False)
    df = client.get(location, scale, year)
    assert df.loc[0, "state_abbreviation"] == location


@pytest.mark.slow
@pytest.mark.parametrize("year", YEARS)
@pytest.mark.parametrize("scale", GEOGRAPHIC_SCALES)
def test_svi_client_get_integration_us(year, scale, loop):
    client = SVIClient(enable_cache=False)
    df = client.get("us", scale, year)
    assert df.state_abbreviation.isin(LOCATIONS).all()



import pytest
from hydrotools.svi_client.types import utilities

# test, validation
LOCATION_TESTS = (
    ("al", "al"),
    ("ALABAMA", "al"),
    ("alabama", "al"),
    ("AlAbAmA", "al"),
)


@pytest.mark.parametrize("test,val", LOCATION_TESTS)
def test_validate_locations(test: str, val: str):
    assert utilities.validate_location(test) == val


GEOGRAPHIC_SCALE_TESTS = (("census_tract", "census_tract"), ("county", "county"))


@pytest.mark.parametrize("test,val", GEOGRAPHIC_SCALE_TESTS)
def test_validate_geographic_scale(test, val):
    assert utilities.validate_geographic_scale(test) == val


GEOGRAPHIC_CONTEXT_TESTS = (("national", "national"), ("state", "state"))


@pytest.mark.parametrize("test,val", GEOGRAPHIC_CONTEXT_TESTS)
def test_validate_geographic_context(test, val):
    assert utilities.validate_geographic_context(test) == val


YEAR_TESTS = (
    (2000, "2000"),
    (2010, "2010"),
    (2014, "2014"),
    (2016, "2016"),
    (2018, "2018"),
    ("2000", "2000"),
    ("2010", "2010"),
    ("2014", "2014"),
    ("2016", "2016"),
    ("2018", "2018"),
)


@pytest.mark.parametrize("test,val", YEAR_TESTS)
def test_validate_year(test, val):
    assert utilities.validate_year(test) == val


# Tests for raising error


def test_validate_locations_raises_ValueError():
    test = "canada"

    with pytest.raises(ValueError):
        utilities.validate_location(test)


def test_validate_geographic_scale_raises_ValueError():
    test = "some fake scale"

    with pytest.raises(ValueError):
        utilities.validate_geographic_scale(test)


def test_validate_geographic_context_raises_ValueError():
    test = "some fake context"

    with pytest.raises(ValueError):
        utilities.validate_geographic_context(test)


def test_validate_year_raises_ValueError():
    test = 1999

    with pytest.raises(ValueError):
        utilities.validate_year(test)



from hydrotools._restclient import Url

# local imports
from .types import utilities, field_name_map
from .types import GeographicScale, GeographicContext, Year
from .consts import US_COUNTY_FEATURE_SERVER_URLS, US_TRACT_FEATURE_SERVERS_URLS

# typing imports
from typing import Optional


def build_csv_url(location: str, geographic_scale: GeographicScale, year: Year) -> str:
    location = utilities.validate_location(location)
    geographic_scale = utilities.validate_geographic_scale(geographic_scale)
    year = utilities.validate_year(year)

    base_path = f"Documents/Data/{year}_SVI_Data/CSV"

    if geographic_scale == "county":
        county_token = get_county_token(year)
        if location == "US":
            location_path = f"SVI{year}_US_{county_token}.csv"
        else:
            location_path = f"States_Counties/{location}_{county_token}.csv"

    else:
        if location == "US":
            location_path = f"SVI{year}_US.csv"
        else:
            location_path = f"States/{location}.csv"

    return f"{base_path}/{location_path}"


def build_feature_server_url(
    location: str,
    geographic_scale: GeographicScale,
    year: Year,
    geographic_context: GeographicContext,
    result_offset: Optional[int] = None,
    result_record_count: Optional[int] = None,
    count_only: bool = False,
) -> str:
    location = utilities.validate_location(location)
    geographic_scale = utilities.validate_geographic_scale(geographic_scale)
    year = utilities.validate_year(year)
    context = utilities.validate_geographic_context(geographic_context)

    if context == "state":
        error_message = (
            "the `state` geographic context has not yet been implimented. "
            "only svi ranked at the `national` context are currently supported."
        )
        raise NotImplemented(error_message)

    path = (
        US_COUNTY_FEATURE_SERVER_URLS[year]
        if geographic_scale == "county"
        else US_TRACT_FEATURE_SERVERS_URLS[year]
    )
    path = f"{path}/query"

    fnm = field_name_map.CdcEsriFieldNameMapFactory(geographic_scale, year)

    params = {
        # for entire US, use 1=1 where clause
        "where": f"{fnm.state_abbreviation} = '{location}'"
        if location != "us"
        else "1=1",
        "outFields": ",".join(
            fnm.dict(exclude_none=True, exclude={"svi_edition"}).values()
        ),
        "returnGeometry": "true",
        "returnExceededLimitFeatures": "true",
        "returnCountOnly": "true" if count_only else "false",
        "resultOffset": "" if result_offset is None else result_offset,
        "resultRecordCount": "" if result_record_count is None else result_record_count,
        "f": "pgeojson",
    }

    o: Url = Url(path, safe="/'") + params
    return o.quote_url


def build_shp_url(location: str, geographic_scale: GeographicScale, year: Year) -> str:
    location = utilities.validate_location(location)
    geographic_scale = utilities.validate_geographic_scale(geographic_scale)
    year = utilities.validate_year(year)

    # shp: Documents/Data/2000_SVI_Data/{geo_option}_2000_SVI.zip
    # US

    base_path = f"Documents/Data/{year}_SVI_Data/CSV"

    if geographic_scale == "county":
        county_token = get_county_token(year)
        if location == "US":
            location_path = f"SVI{year}_US_{county_token}.csv"
        else:
            location_path = f"States_Counties/{location}_{county_token}.csv"

    else:
        if location == "US":
            location_path = f"SVI{year}_US.csv"
        else:
            location_path = f"States/{location}.csv"

    return f"{base_path}/{location_path}"


# helper functions


def get_county_token(year: Year) -> str:
    year = utilities.validate_year(year)

    if year not in ("2014", "2016", "2018"):
        error_message = "Country geographic scale only available for SVI years: 2014, 2016, and 2018."
        raise ValueError(error_message)

    return "CNTY" if year == "2014" else "COUNTY"


# Developer Notes:

# crosswalk urls:
# 2014: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/csv/SVIDocumentation_Table_Crosswalk_2014.csv
# 2016: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/csv/SVIDocumentation_Table_Crosswalk_2016.csv

# data dictionary urls:
# 2000: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/csv/SVIDocumentation_Table_DataDictionary_2000.csv
# 2010: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/csv/SVIDocumentation_Table_DataDictionary_2010.csv
# 2014: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/csv/SVIDocumentation_Table_DataDictionary_2014.csv
# 2016: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/csv/SVIDocumentation_Table_DataDictionary_2016.csv
# 2018: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/csv/SVIDocumentation_Table_DataDictionary_2018.csv

# documentation page urls:
# base_url_plus_path: https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/SVI_documentation_{year}.html

# data dictionary urls:
# 2000: https://svi.cdc.gov/Documents/Data/2000_SVI_Data/SVI2000DataDictionary.pdf
# NOTE: does not appear that 2010 follows the above url convention

# no data value = -999
#
# 14 / 16 / 18 have same col names (in csv case)

# CSVs
# 2000:
#     census_tracts:
#        ✔️ US: Documents/Data/2000_SVI_Data/CSV/SVI2000_US.csv
#        ✔️ single_state: Documents/Data/2000_SVI_Data/CSV/States/{geo_option}.csv
# 2010:
#     census_tracts:
#         ✔️ US: Documents/Data/2010_SVI_Data/CSV/SVI2010_US.csv
#         ✔ single_state: Documents/Data/2010_SVI_Data/CSV/States/{geo_option}.csv
# 2014:
#     census_tracts:
#         US: Documents/Data/2014_SVI_Data/CSV/SVI2014_US.csv
#         single_state: Documents/Data/2014_SVI_Data/CSV/States/{geo_option}.csv
#     counties:
#         US: Documents/Data/2014_SVI_Data/CSV/SVI2014_US_CNTY.csv
#         single_state: Documents/Data/2014_SVI_Data/CSV/States_Counties/{geo_option}_CNTY.csv
# 2016/2018:
#     census_tracts:
#         US: Documents/Data/{year}_SVI_Data/CSV/SVI{year}_US.csv
#         single_state: Documents/Data/{year}_SVI_Data/CSV/States/{geo_option}.csv
#     counties:
#         US: Documents/Data/{year}_SVI_Data/CSV/SVI{year}_US_COUNTY.csv
#         single_state: Documents/Data/{year}_SVI_Data/CSV/States_Counties/{geo_option}_COUNTY.csv


# SHP / GDBs
# 2000s case:
#     census_tracts:
#        US: Documents/Data/2000_SVI_Data/US_2000_SVI.zip # ESRI gdb
#        single_state: https://svi.cdc.gov/Documents/Data/2000_SVI_Data/State_2000_SVI.zip # ESRI gdb

# 2010s case:
#     census_tracts:
#         US: Documents/Data/2010_SVI_Data/SVI2010_US_03242014.zip
#         States: Documents/Data/2010_SVI_Data/SVI2010_States_03242014.zip
#         AllStates: Documents/Data/2010_SVI_Data/SVI2010_States_03242014.zip
#     counties:
#         US: Documents/Data/2010_SVI_Data/SVI2010_Counties.zip
#     unknown_spatial_resolution:
#         single_state: Documents/Data/2010_SVI_Data/States/{geo_option}_03242014.zip

# csv:
#     census_tracts:
#         ✔️ US: Documents/Data/2010_SVI_Data/CSV/SVI2010_US.csv
#         ✔ single_state: Documents/Data/2010_SVI_Data/CSV/States/{geo_option}.csv

# 2014s case:
# shp:
#     census_tracts:
#         US: Documents/Data/2014_SVI_Data/SVI2014_US.zip
#         single_state: Documents/Data/2014_SVI_Data/States/{geo_option).zip
#     counties:
#         US: Documents/Data/2014_SVI_Data/SVI2014_US_CNTY.zip
#         single_state: Documents/Data/2014_SVI_Data/States_Counties/{geo_option}_CNTY.zip

# csv:
#     census_tracts:
#         US: Documents/Data/2014_SVI_Data/CSV/SVI2014_US.csv
#         single_state: Documents/Data/2014_SVI_Data/CSV/States/{geo_option}.csv
#     counties:
#         US: Documents/Data/2014_SVI_Data/CSV/SVI2014_US_CNTY.csv
#         single_state: Documents/Data/2014_SVI_Data/CSV/States_Counties/{geo_option}_CNTY.csv

# 2016s case:
# 2018s case:
# shp:
#     census_tracts:
#         US: Documents/Data/{year}_SVI_Data/SVI{year}_US.zip
#         single_state: Documents/Data/{year}_SVI_Data/States/{geo_option).zip
#     counties:
#         US: Documents/Data/{year}_SVI_Data/SVI{year}_US_COUNTY.zip
#         single_state: Documents/Data/{year}_SVI_Data/States_Counties/{geo_option}_CNTY.zip

# csv:
#     census_tracts:
#         US: Documents/Data/{year}_SVI_Data/CSV/SVI{year}_US.csv
#         single_state: Documents/Data/{year}_SVI_Data/CSV/States/{geo_option}.csv
#     counties:
#         US: Documents/Data/{year}_SVI_Data/CSV/SVI{year}_US_COUNTY.csv
#         single_state: Documents/Data/{year}_SVI_Data/CSV/States_Counties/{geo_option}_COUNTY.csv

# Get counties for a given SVI year
# https://svi.cdc.gov/xmldata/SVI{year}_Counties.xml

# County level maps
#   base url: https://svi.cdc.gov/Documents/CountyMaps/
#
#   example:
#   https://svi.cdc.gov/Documents/CountyMaps/2014/Alabama/Alabama2014_Tuscaloosa.pdf

# 2000:
#   {state}{year}_{county}.pdf

# 2010:
#   {state}{year}_v2_{date-filename | county}.pdf

# 2014 | 2016 | 2018:
#   {state}{year}_{date-filename | county}.pdf
#   https://svi.cdc.gov/Documents/CountyMaps/{year}/{state}/{state}{year}_{data-filename | county}.pdf



from hydrotools._restclient import RestClient
import pandas as pd
import geopandas as gpd

# local imports
from . import url_builders
from .types import GeographicScale, GeographicContext, Year, utilities, field_name_map

# typing imports
from typing import Union
from pathlib import Path


class SVIClient:
    def __init__(
        self,
        enable_cache: bool = True,
        cache_filename: Union[str, Path] = "svi_client_cache",
    ) -> None:
        self._rest_client = RestClient(
            cache_filename=cache_filename,
            enable_cache=enable_cache,
        )

    def get(
        self,
        location: str,
        geographic_scale: GeographicScale,
        year: Year,
        geographic_context: GeographicContext = "national",
    ) -> gpd.GeoDataFrame:
        """Retrieve social vulnerability index thematic rankings and values for a given state or the
        U.S..

        SVI values are available for the following years: 2000, 2010, 2014, 2016, and 2018.  The CDC
        calculates the SVI at the census tract or county geographic scale. Likewise, the CDC
        calculates SVI rankings in two geographic contexts: (1) relative to a given state's SVI
        values or (2) relative to the U.S.. (1) permits interastate comparison and (2) permits
        national comparison.

        Note: `state` geographic_context is not supported at this time.

        Parameters
        ----------
        location : str
            state / national name or abbreviation (e.g. "AL", "US", "Wyoming", "new york")
        geographic_scale : GeographicScale "census_tract" or "county"
            geographic scale at which theme values were calculated
        year : Year
            2000, 2010, 2014, 2016, or 2018
        geographic_context : GeographicContext "national" or "state", optional
            svi rankings calculated at the national or state level. use state for intrastate comparisons, by default "national"
            Note: `state` not supported at this time. will raise NotImplimented Error

        Returns
        -------
        pd.DataFrame
            Dataframe of Social Vulnerability Index values at the census tract or county scale

            columns names:
                state_name: str
                state_abbreviation: str
                county_name: str
                state_fips: str
                county_fips: str
                fips: str
                theme: str
                rank: float
                value: float
                svi_edition: str
                geometry: gpd.array.GeometryDtype


        Examples
        --------
        >>> client = SVIClient()
        ... df = client.get("AL", "census_tract", "2018")
                    state_name state_abbreviation  ... svi_edition                                           geometry
        0        alabama                 al  ...        2018  POLYGON ((-87.21230 32.83583, -87.20970 32.835...
        1        alabama                 al  ...        2018  POLYGON ((-86.45640 31.65556, -86.44864 31.655...
        ...          ...                ...  ...         ...                                                ...
        29498    alabama                 al  ...        2018  POLYGON ((-85.99487 31.84424, -85.99381 31.844...
        29499    alabama                 al  ...        2018  POLYGON ((-86.19941 31.80787, -86.19809 31.808...

        """
        url_path = url_builders.build_feature_server_url(
            location=location,
            geographic_scale=geographic_scale,
            year=year,
            geographic_context=geographic_context,
            count_only=True,
        )

        # RestClient only allows 200 response code or an aiohttp.client_exceptions.ClientConnectorError is raised
        # number of features
        count_request = self._rest_client.get(url_path)

        deserialized_count = count_request.json()
        count = deserialized_count["properties"]["count"]

        # number of features requested by a single request
        OFFSET = 1000
        n_gets = (count // OFFSET) + 1

        urls = [
            url_builders.build_feature_server_url(
                location=location,
                geographic_scale=geographic_scale,
                year=year,
                geographic_context=geographic_context,
                result_offset=i * OFFSET,
                result_record_count=OFFSET,
            )
            for i in range(n_gets)
        ]

        results = self._rest_client.mget(urls)

        # create geodataframe from geojson response
        df = pd.concat(
            [gpd.GeoDataFrame.from_features(r.json()) for r in results],
            ignore_index=True,
        )

        assert len(df) == count

        fnm = field_name_map.CdcEsriFieldNameMapFactory(geographic_scale, year)

        # map of dataset field names to canonical field names
        field_names = {
            v: k
            for k, v in fnm.dict(exclude_none=True, exclude={"svi_edition"}).items()
        }

        df = df.rename(columns=field_names)

        str_cols = df.select_dtypes(include=object).columns
        # lowercase and strip all leading and trailing white spaces from str columns for consistent
        # output and quality control
        df[str_cols] = df[str_cols].apply(lambda d: d.str.strip().str.lower())

        # cast str columns to category type. this reduces the memory footprint by several orders of
        # magnitude
        df[str_cols] = df[str_cols].astype("category")

        # create missing fields if required
        df = fnm.create_missing_fields(df)

        df["svi_edition"] = fnm.svi_edition

        # create column for rejoining dataframes
        df["geometry_idx"] = df.index

        # take a copy of the geometry column
        geometry = df["geometry"]
        df = df.drop(columns=["geometry"])

        # mask of column names that don't end in rank or value
        rank_value_cols_mask = df.columns.str.contains("rank|value$")

        rank_value_col_names = df.columns[rank_value_cols_mask]

        rank_col_names = rank_value_col_names[
            rank_value_col_names.str.contains("rank$")
        ].tolist()

        value_col_names = rank_value_col_names[
            rank_value_col_names.str.contains("value$")
        ].tolist()

        non_rank_value_col_names = df.columns[~rank_value_cols_mask].tolist()

        ranks_df = df.melt(
            id_vars=non_rank_value_col_names,
            value_vars=rank_col_names,
            var_name="rank_theme",
            value_name="rank",
        )

        # some data sources do not include the svi theme values, they only include their rank.
        if value_col_names:
            values_df = df.melt(
                id_vars=non_rank_value_col_names,
                value_vars=value_col_names,
                var_name="value_theme",
                value_name="value",
            )

            ranks_df = ranks_df.set_index(
                non_rank_value_col_names
                + [ranks_df.groupby(non_rank_value_col_names).cumcount()]
            )
            values_df = values_df.set_index(
                non_rank_value_col_names
                + [values_df.groupby(non_rank_value_col_names).cumcount()]
            )

            df = (
                pd.concat([ranks_df, values_df], axis=1)
                # drop groupby cumcount level
                .pipe(
                    lambda d: d.reset_index(level=d.index.nlevels - 1, drop=True)
                ).reset_index()
            )
        else:
            df = ranks_df

        # re-join the geometry column using a shared index and reset the index
        df = pd.concat([df.set_index("geometry_idx"), geometry], axis=1).reset_index(
            drop=True
        )

        # create theme column by truncating rank_theme's _rank suffix
        df["theme"] = df["rank_theme"].str.rstrip("_rank")

        # drop unnecessary cols
        # value_theme column might not exist, so ignore errors when trying to drop
        df = df.drop(columns=["rank_theme", "value_theme"], errors="ignore")

        df.sort_values("state_name", inplace=True, ignore_index=True)

        output_column_order = [
            "state_name",
            "state_abbreviation",
            "county_name",
            "state_fips",
            "county_fips",
            "fips",
            "theme",
            "rank",
            "value",
            "svi_edition",
            "geometry",
        ]

        # reorder dataframe columns
        # note, during reindex, if there are columns not present in dataframe, they will be created
        # with NaN row values
        df = df.reindex(columns=output_column_order)

        return df

    @staticmethod
    def svi_documentation_url(year: Year) -> str:
        year = utilities.validate_year(year)

        urls = {
            "2000": "https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/pdf/SVI2000Documentation-H.pdf",
            "2010": "https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/pdf/SVI-2010-Documentation-H.pdf",
            "2014": "https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/pdf/SVI2014Documentation_01192022.pdf",
            "2016": "https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/pdf/SVI2016Documentation_01192022.pdf",
            "2018": "https://www.atsdr.cdc.gov/placeandhealth/svi/documentation/pdf/SVI2018Documentation_01192022_1.pdf",
        }

        url = urls.get(year, None)

        # raise error if valid year not in urls.
        # when new svi releases are added, this will purposefully break.
        if url is None:
            # raise error
            error_message = (
                f"documentation for year: {year} has not been added to SVIClient."
            )
            raise ValueError(error_message)

        return url
